/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.view

import android.content.Context
import android.content.res.Configuration
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Paint
import android.graphics.Typeface
import android.graphics.drawable.Drawable
import android.graphics.drawable.ShapeDrawable
import android.util.AttributeSet
import android.util.TypedValue
import android.view.MotionEvent
import android.view.View
import android.view.animation.Animation
import android.view.animation.AnimationUtils
import android.widget.AbsListView
import android.widget.ImageView
import android.widget.TextView
import androidx.annotation.ColorInt
import androidx.core.content.ContextCompat
import androidx.core.graphics.drawable.DrawableCompat
import androidx.recyclerview.widget.LinearLayoutManager
import androidx.recyclerview.widget.RecyclerView
import com.kakao.sdk.partner.friend.model.FriendItem
import com.kakao.sdk.partner.friend.util.FilterSearchable
import com.kakao.sdk.partner.friend.util.FriendManager
import com.kakao.sdk.partner.friend.util.PhonemeUtils
import com.kakao.sdk.partner.friend.util.StringUtils.trimZeroWidthChars
import com.kakao.sdk.v2.partner.friend.R
import java.util.*
import kotlin.math.abs
import kotlin.math.max

/**
 * @suppress
 */
internal class SideIndexView : View {

    companion object {
        private const val FAVORITE = "!"
        private const val ETC = "#"
    }

    private val indexItems = ArrayList<String>()
    private var portraitIndexItems: Array<String>
    private var landscapeIndexItems: Array<String>

    private var currentIndex = -1

    private var currentY = -1f

    private val textPaint: Paint = Paint()
    private var textColor: Int = Color.GRAY
    private var indexTextSize: Float = 0.toFloat()

    private var viewHeight: Float = 0.toFloat()

    private var indexItemWidth: Float = 0.toFloat()
    private var indexItemHeight: Float = 0.toFloat()
    private var firstIndexItemY: Float = 0.toFloat()
    private var firstDrawableIndexItemY: Float = 0.toFloat()

    private var favoriteDrawable: Drawable?
    private var touched = false

    private var groupIndexMap: TreeMap<String, Int>? = null
    private var favoriteOffset = -1
    private var extraOffset = -1
    private var friendListOffset: Int = 0
    private var friendListSize: Int = 0

    private var isShowSingleToast: Boolean = false

    private var listScrolling = false
    private var scrolling = false
    private val scrollListener = OnScrollListener()

    private var toastTextView: TextView
    private var toastImageView: ImageView

    var recyclerView: RecyclerView? = null

    val onScrollListener: RecyclerView.OnScrollListener
        get() = scrollListener

    var isEnabledSideIndex = true

    private val sideIndexerComparator: Comparator<String> =
        Comparator { lhs, rhs ->
            FriendManager.compare(
                lhs.trimZeroWidthChars(),
                rhs.trimZeroWidthChars()
            )
        }

    private val sideIndexerOldComparator: Comparator<String> by lazy {
        Comparator<String> { lhs, rhs -> FriendManager.compareForOld(lhs, rhs) }
    }

    constructor(context: Context) : this(context, null)

    constructor(context: Context, attrs: AttributeSet?) : this(context, attrs, 0)

    constructor(context: Context, attrs: AttributeSet?, defStyleAttr: Int) : super(
        context,
        attrs,
        defStyleAttr
    ) {
        context.obtainStyledAttributes(attrs, R.styleable.SideIndexView).run {
            textColor = getColor(R.styleable.SideIndexView_sideindex_text_color, Color.GRAY)
            indexTextSize = getDimension(
                R.styleable.SideIndexView_sideindex_text_size,
                unitValueToPixel(context, TypedValue.COMPLEX_UNIT_SP, 12f).toFloat()
            )
            recycle()
        }

        textPaint.run {
            isAntiAlias = true
            color = textColor
            textSize = indexTextSize
            textAlign = Paint.Align.CENTER
            try {
                typeface = Typeface.SANS_SERIF
            } catch (ignore: Exception) {
            }
        }

        resources.run {
            landscapeIndexItems = getStringArray(R.array.side_indexer_landscape)
            portraitIndexItems = getStringArray(R.array.side_indexer)
            favoriteDrawable =
                ContextCompat.getDrawable(context, R.drawable.kakao_sdk_list_index_ico_favorite)
                    ?.apply {
                        applyTint(this, ContextCompat.getColor(context, R.color.daynight_gray600s))
                    }
            setIternalIndexItems(configuration.orientation)
        }

        singleToast.initializeToast(context, R.layout.kakao_sdk_side_index_single_toast_layout)

        toastTextView = singleToast.findViewById(R.id.toast_tv) as TextView
        toastImageView = singleToast.findViewById(R.id.toast_iv) as ImageView
    }

    fun setDataSource(
        friendList: List<FriendItem>,
        friendListOffset: Int,
        favoriteOffset: Int,
        extraOffset: Int
    ) {
        this.friendListSize = friendList.size
        this.friendListOffset = friendListOffset
        this.favoriteOffset = favoriteOffset
        this.extraOffset = extraOffset

        groupIndexMap = try {
            getGroupIndexMap(friendList, sideIndexerComparator)
        } catch (e: Exception) {
            getGroupIndexMap(friendList, sideIndexerOldComparator)
        }

        invalidate()
    }

    fun setIndexItems(sideIndexRes: Int, sideLandscapeIndexRes: Int) {
        portraitIndexItems = context.resources.getStringArray(sideIndexRes)
        landscapeIndexItems = context.resources.getStringArray(sideLandscapeIndexRes)
        setIternalIndexItems(context.resources.configuration.orientation)
        requestLayout()
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        val height = MeasureSpec.getSize(heightMeasureSpec)

        val fontMetrics = textPaint.fontMetrics
        viewHeight = (height - paddingTop - paddingBottom).toFloat()
        indexItemHeight = viewHeight / indexItems.size

        for (indexItem in indexItems) {
            indexItemWidth = max(indexItemWidth, textPaint.measureText(indexItem))
        }

        super.onMeasure(
            MeasureSpec.makeMeasureSpec(
                (indexItemWidth + paddingLeft + paddingRight).toInt(),
                MeasureSpec.EXACTLY
            ), heightMeasureSpec
        )

        firstIndexItemY =
            (height - viewHeight) / 2 + (indexItemHeight / 2 - (fontMetrics.descent - fontMetrics.ascent) / 2) - fontMetrics.ascent

        firstDrawableIndexItemY =
            (paddingTop + (indexItemHeight / 2 - (fontMetrics.descent - fontMetrics.ascent) / 2))
    }


    override fun onDraw(canvas: Canvas) {
        super.onDraw(canvas)
        for (i in indexItems.indices) {
            val baseLineY = firstIndexItemY + indexItemHeight * i
            val scale = calcItemScale(i)
            textPaint.let { paint ->
                paint.alpha = if (i == currentIndex) 255 else (255 * (1 - scale)).toInt()
                paint.textSize = indexTextSize + indexTextSize * scale * 0.5f
            }
            val baseLineX = paddingLeft + indexItemWidth / 2

            if (indexItems[i] == FAVORITE) {
                favoriteDrawable?.run {
                    val top = (firstDrawableIndexItemY + indexItemHeight * i).toInt()
                    setBounds(
                        paddingLeft,
                        top,
                        paddingLeft + indexItemWidth.toInt(),
                        top + indexItemWidth.toInt()
                    )
                    draw(canvas)
                }
            } else {
                canvas.drawText(indexItems[i], baseLineX, baseLineY, textPaint)
            }
        }

        textPaint.run {
            alpha = 255
            textSize = indexTextSize
        }
    }

    override fun onDetachedFromWindow() {
        super.onDetachedFromWindow()
        singleToast.destroy()
    }

    override fun setVisibility(visibility: Int) {
        setVisibility(visibility, true)
    }

    override fun onTouchEvent(event: MotionEvent): Boolean {
        if (indexItems.isEmpty() || visibility != VISIBLE) {
            return super.onTouchEvent(event)
        }

        currentIndex = getSelectedIndex(event.y)
        when (event.action) {
            MotionEvent.ACTION_DOWN -> {
                touched = true
                invalidate()
                //this.onTouchListener?.onTouchDown()
                return true
            }
            MotionEvent.ACTION_MOVE -> {
                scrolling = true
                if (touched) {
                    scrollToPosition(indexItems[currentIndex])
                }
                invalidate()
                return true
            }

            MotionEvent.ACTION_UP, MotionEvent.ACTION_CANCEL -> {
                currentIndex = -1
                touched = false
                invalidate()
                //this.onTouchListener?.onTouchUp()
                return true
            }
        }

        return super.onTouchEvent(event)
    }

    private fun setVisibility(visibility: Int, usingAnim: Boolean) {
        if (usingAnim) {
            var anim: Animation? = null
            when (visibility) {
                INVISIBLE, GONE -> anim =
                    AnimationUtils.loadAnimation(context, R.anim.kakao_sdk_fade_out_short)
                VISIBLE -> anim =
                    AnimationUtils.loadAnimation(context, R.anim.kakao_sdk_fade_in_short)
            }
            this.animation = anim
            this.startAnimation(anim)
        } else {
            singleToast.removeToast(0)
        }

        if (visibility != VISIBLE) {
            currentIndex = -1
        }
        super.setVisibility(visibility)
    }

    private inner class OnScrollListener : RecyclerView.OnScrollListener() {
        var scrollState = RecyclerView.SCROLL_STATE_IDLE
        override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
            if (isEnabledSideIndex && !needScroll() && !listScrolling) {
                return
            }
            when (newState) {
                RecyclerView.SCROLL_STATE_DRAGGING, RecyclerView.SCROLL_STATE_SETTLING -> if (!listScrolling && scrollState != newState && friendListSize > 0) {
                    listScrolling = true
                    visibility = VISIBLE
                }
                RecyclerView.SCROLL_STATE_IDLE -> if (listScrolling && !scrolling) {
                    postDelayed({
                        if (scrolling) {
                            return@postDelayed
                        }
                        listScrolling = false
                        if (visibility >= INVISIBLE) {
                            return@postDelayed
                        }
                        visibility = INVISIBLE
                    }, 1000)
                }
            }
            scrollState = newState
        }
    }

    private fun needScroll(): Boolean = recyclerView?.run {
        computeVerticalScrollExtent() < computeVerticalScrollRange()
    } ?: false

    private fun getSelectedIndex(eventY: Float): Int {
        currentY = eventY - (height / 2 - viewHeight / 2)
        if (currentY <= 0) {
            return 0
        }
        var index = (currentY / indexItemHeight).toInt()
        if (index >= indexItems.size) {
            index = indexItems.size - 1
        }
        return index
    }

    private fun scrollToPosition(index: String) {
        val position = getHigherValue(groupIndexMap, index)
        if (index == FAVORITE) {
            (recyclerView?.layoutManager as LinearLayoutManager).scrollToPositionWithOffset(
                favoriteOffset,
                0
            )
        } else if (position == null) {
            recyclerView?.let { recyclerView ->
                recyclerView.adapter?.let { adapter ->
                    recyclerView.scrollToPosition(adapter.itemCount - 1)
                }
            }
        } else {
            if (recyclerView?.layoutManager is LinearLayoutManager) {
                (recyclerView?.layoutManager as LinearLayoutManager).scrollToPositionWithOffset(
                    position + favoriteOffset + extraOffset,
                    0
                )
            } else {
                recyclerView?.scrollToPosition(position + friendListOffset)
            }
        }

        if (index == FAVORITE) {
            toastImageView.apply {
                setImageDrawable(
                    applyTint(
                        ContextCompat.getDrawable(
                            context,
                            R.drawable.kakao_sdk_list_index_ico_favorite_large
                        ),
                        ContextCompat.getColor(context, R.color.daynight_gray600s)
                    )
                )
            }.visibility = VISIBLE
            toastTextView.visibility = GONE
        } else {
            toastTextView.apply {
                text = index
            }.visibility = VISIBLE
            toastImageView.visibility = GONE
        }

        isShowSingleToast = true
        singleToast.setVisibility(VISIBLE)
        singleToast.removeToast(1500)
    }

    private fun getGroupIndexMap(
        friendList: List<FriendItem>,
        comparator: Comparator<String>
    ): TreeMap<String, Int> {
        val alphabetIndexSet: TreeSet<String> by lazy {
            val alphabetIndexSet = TreeSet(comparator)
            for (alphabet in 'a'..'z') {
                alphabetIndexSet.add(alphabet.toString())
            }
            alphabetIndexSet
        }
        val index = TreeMap<String, Int>(comparator)
        var prevChar: Char = 0.toChar()

        for (i in friendList.indices) {
            val o = friendList[i]
            val curChar: Char = if (o is FilterSearchable) {
                PhonemeUtils.getFirstPhoneme(o.filterKeyword.trimZeroWidthChars())
            } else {
                prevChar
            }

            if (prevChar == curChar) {
                continue
            }

            val groupName = curChar.toString()
            if (groupName.length > 1) {
                continue
            }

            val key = when (PhonemeUtils.startsWith(curChar)) {
                PhonemeUtils.ALPHABET -> {
                    try {
                        alphabetIndexSet.floor(curChar.toString()).toUpperCase(Locale.ENGLISH)
                    } catch (e: Exception) {
                        groupName.toUpperCase(FriendManager.comparator.locale())
                    }
                }
                PhonemeUtils.UNKNOWN -> {
                    ETC
                }
                else -> {
                    groupName.toUpperCase(FriendManager.comparator.locale())
                }
            }

            if (!index.containsKey(key)) {
                index[key] = i
            }
            prevChar = curChar
        }

        return index
    }

    private val singleToast = object : SingleToast() {

        override fun removeWindow() {
            if (isShowSingleToast) {
                isShowSingleToast = false
                scrolling = false
                this.setVisibility(INVISIBLE)
                scrollListener.onScrollStateChanged(
                    recyclerView!!,
                    AbsListView.OnScrollListener.SCROLL_STATE_IDLE
                )
            }
        }
    }

    private fun calcItemScale(index: Int): Float {
        var scale = 0f
        if (currentIndex != -1) {
            val distance =
                abs(currentY - (indexItemHeight * index + indexItemHeight / 2)) / indexItemHeight
            scale = max(1 - distance * distance / 16, 0f)
        }
        return scale
    }

    private fun setIternalIndexItems(orientation: Int) {
        indexItems.clear()
        if (favoriteOffset != -1) {
            indexItems.add(FAVORITE)
        }

        if (orientation == Configuration.ORIENTATION_PORTRAIT) {
            for (portraitIndexItem in portraitIndexItems) {
                indexItems.add(portraitIndexItem)
            }
        } else {
            for (landscapeIndexItem in landscapeIndexItems) {
                indexItems.add(landscapeIndexItem)
            }
        }
    }

    private fun <K, V> getHigherValue(map: TreeMap<K, V>?, key: K?): V? {
        if (map == null || key == null) {
            return null
        }
        try {
            val firstKey = map.firstKey()
            val lastKey = map.lastKey()

            if (lastKey == key) {
                return map[lastKey]
            }

            if (firstKey == key) {
                return map[firstKey]
            }

            val sortedMap = map.tailMap(key)

            return if (sortedMap.size == 0) {
                null
            } else sortedMap[sortedMap.firstKey()]

        } catch (nse: NoSuchElementException) {
            //Logger.wt(Logger.Tag.EXPANDABLE, "", nse)
        }

        return null
    }

    private fun unitValueToPixel(context: Context, unit: Int, unitValue: Float): Int {
        return TypedValue.applyDimension(unit, unitValue, context.resources.displayMetrics).toInt()
    }

    private fun applyTint(drawable: Drawable?, @ColorInt tintColor: Int): Drawable {
        if (drawable == null) return ShapeDrawable()
        val wrapped = DrawableCompat.wrap(drawable)
        DrawableCompat.setTint(wrapped.mutate(), tintColor)
        return wrapped
    }
}